from numpy import dot

from .base import BaseStepAssociative


__all__ = ('Instar',)


class Instar(BaseStepAssociative):
    """ Instar is a simple unsupervised Neural Network algortihm which
    detect associations in features. Algorithm very similar to
    :network:`HebbRule` except the learning rule.

    Parameters
    ----------
    n_unconditioned : int
        This value control number of features which are unconditioned
        stimulus for network. Defaults to ``1``. Can be any integer value
        bigger than ``1``, but less than feature space.
    {step}
    {show_epoch}
    {shuffle_data}
    {error}
    {verbose}
    {full_signals}

    Methods
    -------
    {unsupervised_train_epochs}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy import algorithms, layers
    >>>
    >>> train_data = np.array([
    ...     [0, 1, 0, 0],
    ...     [1, 1, 0, 0],
    ... ])
    >>> test_cases = np.array([
    ...     [0, 1, 0, 0],
    ...     [0, 0, 0, 0],
    ...     [0, 0, 1, 1],
    ... ])
    >>>
    >>>
    >>> scaler = {{'lower_value': 0, 'upper_value': 1}}
    >>> input_layer = layers.StepLayer(4, function_coef=scaler)
    >>> output_layer = layers.OutputLayer(1)
    >>>
    >>> instnet = algorithms.Instar(
    ...     input_layer > output_layer,
    ...     n_unconditioned=1,
    ...     step=1,
    ...     verbose=False,
    ... )
    >>>
    >>> instnet.train(train_data, epochs=2)
    >>> instnet.predict(test_cases)
    array([[1],
           [0],
           [0]])
    """
    def weight_delta(self, input_row, layer_output):
        unconditioned = self.n_unconditioned
        update_from_column = unconditioned - self.use_bias
        weight = self.input_layer.weight[unconditioned:, :]
        return self.step * dot(
            (input_row[:, update_from_column:].T - weight), layer_output
        )



from math import sqrt, pi

from numpy import zeros, tile, exp
from numpy.core.umath_tests import inner1d


__all__ = ('pdf_between_data',)


def pdf_between_data(train_data, input_data, standard_deviation):
    """ Compute PDF between two samples.

    Parameters
    ----------
    train_data : array
        train sample
    input_data : array
        input sample
    standard_deviation : float
        standard deviation for PDF
    """
    # Note: This implementation works faster than 3D arrays
    # and use less memory.
    results = zeros((train_data.shape[0], input_data.shape[0]))
    variance = standard_deviation ** 2
    function_const = standard_deviation * sqrt(2 * pi)
    train_data_size = train_data.shape[0]

    for i, input_row in enumerate(input_data):
        inputs = tile(input_row, (train_data_size, 1))
        class_difference = (train_data - inputs)
        total_distance = inner1d(class_difference, class_difference)
        results[:, i] = exp(-total_distance / variance) / function_const

    return results



import os

from setuptools import setup, find_packages

import neuralpy


CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))


def get_requirements():
    try:
        requirements_file = os.path.join(CURRENT_DIR, 'requirements/main.txt')
        with open(requirements_file) as f:
            return f.read().splitlines()
    except IOError:
        # Simple hack for `tox` test.
        return []


setup(
    # Info
    name='neupy',
    version=neuralpy.__version__,
    license='MIT',
    url='http://neuralpy.com',
    description=neuralpy.__doc__,

    # Author
    author='Yurii Shevhcuk',
    author_email='mail@itdxer.com',

    # Package
    packages=find_packages(),
    install_requires=get_requirements(),
    include_package_data=True,
    zip_safe=False,
    keywords=['neural networks'],

    # Command
    entry_points={
        'console_scripts': ['neuralpy = neuralpy.commands.main:main'],
    },

    # Other
    setup_cfg=True,
    classifiers=[
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import numpy as np
from sklearn import datasets
from sklearn.cross_validation import StratifiedKFold

from neuralpy.algorithms import PNN


dataset = datasets.load_iris()
data = dataset.data
target = dataset.target

test_data_size = 10
skfold = StratifiedKFold(target, test_data_size)
avarage_result = 0

print("> Start classify iris dataset")

for i, (train, test) in enumerate(skfold, start=1):
    x_train, x_test = data[train], data[test]
    y_train, y_test = target[train], target[test]

    pnn_network = PNN(standard_deviation=0.1, verbose=False)
    pnn_network.train(x_train, y_train)
    result = pnn_network.predict(x_test)
    print("Test #{:<2}: {}/{}".format(
        i, np.sum(result == y_test), test.size
    ))



from itertools import chain

from numpy import dot

from neuralpy.core.properties import ListProperty
from neuralpy.algorithms.feedforward import FeedForwardNetwork
from neuralpy.network.learning import SupervisedLearning
from . import optimization_types


__all__ = ('Backpropagation',)


class Backpropagation(SupervisedLearning, FeedForwardNetwork):
    """ Backpropagation algorithm.

    Parameters
    ----------
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms import Backpropagation
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> bpnet = Backpropagation((2, 3, 1), verbose=False, step=0.1)
    >>> bpnet.train(x_train, y_train)
    """

    __opt_params = """optimizations : list or None
        The list of optimization algortihms. ``None`` by default.
        If this option is not empty it will generate new class which
        will inherit all from this list. Support two types of
        optimization algorithms: weight update and step update.
    """
    shared_docs = {"optimizations": __opt_params}

    optimizations = ListProperty(default=None)
    default_optimizations = []

    def __new__(cls, connection, options=None, **kwargs):
        # Argument `options` is a simple hack for the `__reduce__` method.
        # `__reduce__` can't retore class with keyword arguments and
        # it will put them as `dict` argument in the `options` and method
        # will translate it to kwargs. The same hack is at the
        # `__init__` method.
        if options is None:
            options = kwargs

        optimizations = options.get('optimizations', cls.default_optimizations)
        if not optimizations:
            return super(Backpropagation, cls).__new__(cls)

        founded_types = []
        for optimization_class in optimizations:
            opt_class_type = getattr(optimization_class, 'optimization_type',
                                     None)
            if opt_class_type not in optimization_types:
                raise ValueError("Invalid optimization class `{}`".format(
                                 optimization_class.__name__))

            if opt_class_type in founded_types:
                raise ValueError(
                    "There can be only one optimization class with "
                    "type `{}`".format(optimization_types[opt_class_type])
                )

            founded_types.append(opt_class_type)

        # Build new class which would inherit main and all optimization
        new_class_name = (
            cls.__name__ +
            ''.join(class_.__name__ for class_ in optimizations)
        )
        mro_classes = tuple(list(optimizations) + [cls])
        new_class = type(new_class_name, mro_classes, {})

        return super(Backpropagation, new_class).__new__(new_class)

    def __init__(self, connection, options=None, **kwargs):
        if options is None:
            options = kwargs

        default_optimizations = self.default_optimizations
        optimizations = options.get('optimizations', default_optimizations)

        if optimizations != default_optimizations:
            optimizations_merged = []
            for algorithm in chain(optimizations, default_optimizations):
                types = [alg.optimization_type for alg in optimizations_merged]
                if algorithm.optimization_type not in types:
                    optimizations_merged.append(algorithm)
        else:
            optimizations_merged = optimizations

        options['optimizations'] = optimizations_merged
        super(Backpropagation, self).__init__(connection, **options)

    def get_gradient(self, output_train, target_train):
        self.delta = []
        self.gradients = []

        update = self.error.deriv(output_train, target_train)

        for i, layer in enumerate(reversed(self.train_layers), start=1):
            summated_data = self.summated_data[-i]
            current_layer_input = self.layer_outputs[-i]

            deriv = layer.activation_function.deriv(summated_data)

            delta = deriv * update
            update = dot(delta, layer.weight_without_bias.T)

            gradient = dot(current_layer_input.T, delta)

            self.gradients.insert(0, gradient)
            self.delta.insert(0, delta)

        return self.gradients

    def get_weight_delta(self, output_train, target_train):
        gradients = self.get_gradient(output_train, target_train)
        return [-gradient for gradient in gradients]

    def update_weights(self, weight_deltas):
        layer_weight_update = self.layer_weight_update
        for i, layer in enumerate(self.train_layers):
            layer.weight += layer_weight_update(weight_deltas[i], i)

    def layer_step(self, layer_number):
        return self.step

    def layer_weight_update(self, delta, layer_number):
        return self.layer_step(layer_number) * delta

    def get_class_name(self):
        return 'Backpropagation'

    def __reduce__(self):
        options = {}
        for name, option in self.options.items():
            value = getattr(self, name)
            # Default values not always valid types. For this reason we
            # ignore all values which has the same as default.
            if value != option.value.default:
                options[name] = value
        return (Backpropagation, (self.connection, options))



from abc import ABCMeta
from collections import namedtuple
from functools import reduce

from six import with_metaclass

from .properties import Property
from .docs import docs


__all__ = ('ConfigMeta', 'ConfigWithABCMeta', 'Configurable',
           'ConfigurableWithABC')


Option = namedtuple('Option', 'class_name value')


def merge_dicts(left_dict, right_dict):
    return dict(left_dict, **right_dict)


class ConfigMeta(type):
    def __new__(cls, clsname, bases, attrs):
        parents = [kls for kls in bases if isinstance(kls, ConfigMeta)]
        new_class = super(ConfigMeta, cls).__new__(cls, clsname, bases, attrs)

        if new_class.__doc__ is not None:
            maindocs = docs.copy()
            mro_classes = new_class.__mro__

            # Collect parameter `shared_docs` for all MRO classes and
            # combine them in one big dictionary
            shared_docs = [getattr(b, 'shared_docs', {}) for b in mro_classes]
            all_params = reduce(merge_dicts, shared_docs, maindocs)

            new_class.__doc__ = new_class.__doc__.format(**all_params)

        if not hasattr(new_class, 'options'):
            new_class.options = {}

        # Populate parent classes options
        for base_class in parents:
            if hasattr(base_class, 'options'):
                new_class.options = dict(base_class.options,
                                         **new_class.options)

        # Set properties names and save options for different classes
        for key, value in attrs.items():
            if isinstance(value, Property):
                value.name = key
                new_class.options[key] = Option(
                    class_name=clsname,
                    value=value,
                )

        return new_class


class ConfigWithABCMeta(ABCMeta, ConfigMeta):
    pass


class BaseConfigurable(object):
    def __init__(self, **options):
        available_options = set(self.options.keys())
        invalid_options = set(options) - available_options

        if invalid_options:
            raise ValueError("Network `{}` contains invalid properties: "
                             "{}".format(self.__class__.__name__,
                                         ', '.join(invalid_options)))

        for key, value in options.items():
            setattr(self, key, value)


class Configurable(with_metaclass(ConfigMeta, BaseConfigurable)):
    pass


class ConfigurableWithABC(with_metaclass(ConfigWithABCMeta,
                                         BaseConfigurable)):
    pass



from operator import mul

from numpy import ogrid, reshape, nonzero

from neuralpy.core.properties import (NonNegativeIntProperty,
                                      NumberBoundProperty)
from neuralpy.layers import CompetitiveOutputLayer
from neuralpy.algorithms import Kohonen


__all__ = ('SOFM',)


def neuron_neighbours(neurons, center, radius):
    """ Function find all neighbours neurons by radius and coords.

    Parameters
    ----------
    neurons : arary-like
        Array element with neurons.
    center : tuple
        Index of the main neuron for which function must find neighbours.
    radius : int
        Radius indetify which neurons hear the main one are neighbours.

    Returns
    -------
    array-like
        Return matrix with the same dimention as ``neurons`` where center
        element and it neighbours positions filled with value ``1``
        and other as ``0``.

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms.competitive.sofm import neuron_neighbours
    >>>
    >>> neuron_neighbours(np.zeros((3, 3)), (0, 0), 1)
    array([[ 1.,  1.,  0.],
           [ 1.,  0.,  0.],
           [ 0.,  0.,  0.]])
    >>> neuron_neighbours(np.zeros((5, 5)), (2, 2), 2)
    array([[ 0.,  0.,  1.,  0.,  0.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 1.,  1.,  1.,  1.,  1.],
           [ 0.,  1.,  1.,  1.,  0.],
           [ 0.,  0.,  1.,  0.,  0.]])
    """
    center_x, center_y = center
    nx, ny = neurons.shape

    y, x = ogrid[-center_x:nx - center_x, -center_y:ny - center_y]
    mask = (x * x + y * y) <= radius ** 2
    neurons[mask] = 1

    return neurons


class SOFM(Kohonen):
    """ Self-Organizing Feature Map.

    Notes
    -----
    * Network architecture must contains two layers.
    * Second layer must be :layer:`CompetitiveOutputLayer`.

    Parameters
    ----------
    learning_radius : int
        Learning radius.
    features_grid : int
        Learning radius.
    {full_params}

    Methods
    -------
    {unsupervised_train_epochs}
    {predict}
    {plot_errors}
    {last_error}
    """
    learning_radius = NonNegativeIntProperty(default=0)
    features_grid = NumberBoundProperty()
    # # None - mean that this property is the same as default step
    # neighbours_step = NumberProperty()

    def __init__(self, connection, **options):
        super(SOFM, self).__init__(connection, **options)

        if not isinstance(self.output_layer, CompetitiveOutputLayer):
            raise ValueError("Output layer must be `CompetitiveOutputLayer`")

        if self.features_grid is not None:
            if mul(*self.features_grid) != self.output_layer.input_size:
                raise ValueError(
                    "Feature grid must contains the same size of elements as "
                    "at output layer: {0}. But it contains: {1} "
                    "({2}x{3})".format(
                        self.output_layer.input_size,
                        mul(*self.features_grid),
                        self.features_grid[0],
                        self.features_grid[1]
                    )
                )

    def setup_defaults(self):
        super(SOFM, self).setup_defaults()

        # if self.neighbours_step is None:
        #     self.neighbours_step = self.step

        if self.features_grid is None:
            self.features_grid = (self.output_layer.input_size, 1)

    def update_indexes(self, layer_output):
        neuron_winner = layer_output.argmax(axis=1)
        feature_bound = self.features_grid[1]

        output_with_neightbours = neuron_neighbours(
            reshape(layer_output, self.features_grid),
            (neuron_winner // feature_bound, neuron_winner % feature_bound),
            self.learning_radius
        )
        index_y, _ = nonzero(
            reshape(
                output_with_neightbours,
                (self.output_layer.input_size, 1)
            )
        )
        return index_y



from __future__ import division

from neuralpy.core.properties import NonNegativeIntProperty, NumberProperty
from .base import SingleStep


__all__ = ('SearchThenConverge',)


class SearchThenConverge(SingleStep):
    """ Algorithm minimize learning step. Similar to
    :network:`SimpleStepMinimization`, but more complicated step update rule.

    Parameters
    ----------
    epochs_step_minimizator : int
        The parameter controls the frequency reduction step with respect
        to epochs. Defaults to ``100`` epochs. Can't be less than ``1``.
        Less value mean that step decrease faster.
    rate_coefitient : float
        Second important parameter to control the rate of error reduction.
        Defaults to ``0.2``

    Attributes
    ----------
    {first_step}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> from neuralpy import algorithms
    >>>
    >>> bpnet = algorithms.Backpropagation(
    ...     (2, 4, 1),
    ...     step=0.1,
    ...     verbose=False,
    ...     optimizations=[algorithms.SearchThenConverge]
    ... )
    >>>

    See Also
    --------
    :network:`SimpleStepMinimization`
    """
    epochs_step_minimizator = NonNegativeIntProperty(min_size=1, default=100)
    rate_coefitient = NumberProperty(default=0.2)

    def after_weight_update(self, input_train, target_train):
        super(SearchThenConverge, self).after_weight_update(
            input_train, target_train
        )

        first_step = self.first_step
        epochs_step_minimizator = self.epochs_step_minimizator

        epoch_value = self.epoch / epochs_step_minimizator
        rated_value = (self.rate_coefitient / first_step) * epoch_value

        self.step = first_step * (1 + rated_value) / (
            1 + rated_value + epochs_step_minimizator * epoch_value ** 2
        )



import types

import numpy as np

from neuralpy.helpers.base import preformat_value


__all__ = ('Property', 'CheckSizeProperty', 'NumberProperty', 'BoolProperty',
           'BetweenZeroAndOneProperty', 'FuncProperty', 'NumberBoundProperty',
           'ArrayProperty', 'ListOfTypesProperty', 'ListProperty',
           'DictProperty', 'IntProperty', 'StringProperty',
           'NonNegativeIntProperty', 'NonNegativeNumberProperty',
           'ChoiceProperty')


class Property(object):
    expected_type = object
    disable = True

    def __init__(self, name=None, default=None, **options):
        self.name = name
        self.default = default

        for key, value in options.items():
            setattr(self, key, value)

    def __set__(self, instance, value):
        if not isinstance(value, self.expected_type):
            availabe_types = self.expected_type
            if isinstance(availabe_types, (list, tuple)):
                availabe_types = ', '.join(t.__name__ for t in availabe_types)
            else:
                availabe_types = availabe_types.__name__

            raise TypeError(
                "Wrong data type `{0}` for `{1}` property. Expected types: "
                "{2}".format(value.__class__.__name__, self.name,
                             availabe_types)
            )

        self.validate(value)
        instance.__dict__[self.name] = value

    def __get__(self, instance, value):
        if instance is None:
            return

        if self.default is not None and self.name not in instance.__dict__:
            self.__set__(instance, self.default)

        return instance.__dict__.get(self.name, None)

    def __delete__(self, instance):
        name = self.name

        if name in instance.__dict__:
            del instance.__dict__[name]

        if name in instance.options:
            del instance.options[name]

    def __str__(self):
        return str(preformat_value(self.default))

    def __repr__(self):
        return self.__str__()

    def validate(self, value):
        pass


# -----------------------------------------------------#
#                   Typed properties                   #
# -----------------------------------------------------#


class StringProperty(Property):
    expected_type = str


class IntProperty(Property):
    expected_type = int


class NumberProperty(Property):
    expected_type = (float, int)


class BoolProperty(Property):
    expected_type = bool


class DictProperty(Property):
    expected_type = dict


class FuncProperty(Property):
    expected_type = types.FunctionType


class ArrayProperty(Property):
    expected_type = np.ndarray


class ListProperty(Property):
    expected_type = list


# -----------------------------------------------------#
#                 Special properties                   #
# -----------------------------------------------------#


class ListOfTypesProperty(Property):
    expected_type = (list, tuple, set)
    inner_list_type = int
    count = None

    def validate(self, value):
        super(ListOfTypesProperty, self).validate(value)

        if self.count is not None and len(value) != self.count:
            raise ValueError("Expected list with {} variables".format(
                self.count
            ))

        if not all(isinstance(v, self.inner_list_type) for v in value):
            raise TypeError("Expected list with {}".format(
                self.inner_list_type.__name__
            ))


class NumberBoundProperty(ListOfTypesProperty):
    count_of_values = 2


class VectorProperty(Property):
    expected_type = np.ndarray

    def validate(self, value):
        if value.ndim != 1:
            raise ValueError(
                "Value `{}` must be 1-D shape vector".format(self.name)
            )


class Matrix2DProperty(Property):
    expected_type = (np.ndarray, np.matrix)

    def validate(self, value):
        if value.ndim != 2:
            raise ValueError(
                "Value `{}` must be 2-D shape array/matrix".format(self.name)
            )


# -----------------------------------------------------#
#                  Choices properties                  #
# -----------------------------------------------------#


class ChoiceProperty(Property):
    choices = {}

    def __init__(self, *args, **kwargs):
        super(ChoiceProperty, self).__init__(*args, **kwargs)
        choices = self.choices

        if isinstance(choices, (list, tuple)):
            self.choices = dict(zip(choices, choices))

        if not isinstance(self.choices, dict):
            raise ValueError("Choice properties can be only a `dict`, got "
                             "`{0}`".format(self.choices.__class__.__name__))

        if not self.choices:
            raise ValueError("Must be at least one choice in property "
                             "`{0}`".format(self.name))

    def __set__(self, instance, value):
        if value not in self.choices:
            raise ValueError(
                "Wrong value `{0}` for property `{1}`. Available values: "
                "{2}".format(value, self.name, ", ".join(self.choices.keys()))
            )
        return super(ChoiceProperty, self).__set__(instance, value)

    def __get__(self, instance, value):
        if instance is not None:
            choice_key = super(ChoiceProperty, self).__get__(instance, value)
            return self.choices[choice_key]


# -----------------------------------------------------#
#                   Sized properties                   #
# -----------------------------------------------------#


class CheckSizeProperty(Property):
    min_size = -np.inf
    max_size = np.inf

    def __init__(self, min_size=None, max_size=None, *args, **kwargs):
        if min_size is not None:
            self.min_size = min_size

        if max_size is not None:
            self.max_size = max_size

        super(CheckSizeProperty, self).__init__(*args, **kwargs)

    def validate(self, value):
        if not self.min_size <= value <= self.max_size:
            raise ValueError("Value `{}` must be between {} and {}".format(
                self.name, self.min_size, self.max_size
            ))


class BetweenZeroAndOneProperty(NumberProperty, CheckSizeProperty):
    min_size = 0
    max_size = 1


class NonNegativeIntProperty(IntProperty, CheckSizeProperty):
    min_size = 0


class NonNegativeNumberProperty(NumberProperty, CheckSizeProperty):
    min_size = 0



from __future__ import division

import copy

from numpy import where, sign, abs as np_abs

from neuralpy.core.properties import NonNegativeNumberProperty
from .backpropagation import Backpropagation


__all__ = ('Quickprop',)


class Quickprop(Backpropagation):
    """ Quickprop :network:`Backpropagation` algorithm optimization.

    Parameters
    ----------
    upper_bound : float
        Maximum possible value for weight update. Defaults to ``1``.
    {optimizations}
    {raw_predict_param}
    {full_params}

    Methods
    -------
    {supervised_train}
    {full_methods}

    Examples
    --------
    Simple example

    >>> import numpy as np
    >>> from neuralpy import algorithms
    >>>
    >>> x_train = np.array([[1, 2], [3, 4]])
    >>> y_train = np.array([[1], [0]])
    >>>
    >>> qpnet = algorithms.Quickprop(
    ...     (2, 3, 1),
    ...     verbose=False
    ... )
    >>> qpnet.train(x_train, y_train)

    See Also
    --------
    :network:`Backpropagation` : Backpropagation algorithm.
    """
    upper_bound = NonNegativeNumberProperty(default=1)

    def layer_weight_update(self, delta, layer_number):
        if not hasattr(self, 'prev_gradients'):
            weight_delta = delta
        else:
            gradient = self.gradients[layer_number]
            prev_gradient = self.prev_gradients[layer_number]
            prev_weight_delta = self.prev_weight_deltas[layer_number]
            weight_delta = prev_weight_delta * (
                gradient / (prev_gradient - gradient)
            )
            upper_bound = self.upper_bound
            weight_delta = where(
                np_abs(weight_delta) < upper_bound, weight_delta,
                sign(weight_delta) * upper_bound
            )

        self.weight_deltas.append(weight_delta)
        return weight_delta

    def update_weights(self, weight_deltas):
        self.weight_deltas = []
        super(Quickprop, self).update_weights(weight_deltas)

        self.prev_weight_deltas = copy.copy(self.weight_deltas)
        self.prev_gradients = copy.copy(self.gradients)



from numpy import dot

from neuralpy.core.properties import NonNegativeNumberProperty
from neuralpy.network.base import BaseNetwork
from neuralpy.network.connections import FAKE_CONNECTION
from neuralpy.network.learning import LazyLearning
from neuralpy.network.types import Regression
from .utils import pdf_between_data


__all__ = ('GRNN',)


class GRNN(LazyLearning, Regression, BaseNetwork):
    """ Generalized Regression Neural Network.

    Parameters
    ----------
    standard_deviation : float
        standard deviation for PDF function, default to 0.1.
    {full_params}

    Methods
    -------
    {supervised_train_lazy}
    {full_methods}

    Examples
    --------
    >>> import numpy as np
    >>>
    >>> from sklearn import datasets
    >>> from sklearn.cross_validation import train_test_split
    >>>
    >>> from neuralpy.algorithms import GRNN
    >>> from neuralpy.functions import rmsle
    >>>
    >>> np.random.seed(0)
    >>>
    >>> dataset = datasets.load_diabetes()
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     dataset.data, dataset.target, train_size=0.7,
    ...     random_state=0
    ... )
    >>>
    >>> nw = GRNN(standard_deviation=0.1, verbose=False)
    >>> nw.train(x_train, y_train)
    >>> result = nw.predict(x_test)
    >>> rmsle(result, y_test)
    0.4245120142774001
    """
    standard_deviation = NonNegativeNumberProperty(default=0.1)

    def __init__(self, **options):
        super(GRNN, self).__init__(FAKE_CONNECTION, **options)

    def train(self, input_train, target_train):
        if target_train.ndim != 1:
            raise ValueError("Target value must be in 1 dimention")
        LazyLearning.train(self, input_train, target_train)

    def predict(self, input_data):
        super(GRNN, self).predict(input_data)

        input_data_size = input_data.shape[1]
        train_data_size = self.input_train.shape[1]

        if input_data_size != train_data_size:
            raise ValueError("Input data must contains {0} features, got "
                             "{1}".format(train_data_size, input_data_size))

        ratios = pdf_between_data(self.input_train, input_data,
                                  self.standard_deviation)
        return dot(ratios.T, self.target_train) / ratios.sum(axis=0)



import numpy as np

from neuralpy import layers, algorithms


input_data = np.array([
    [-0.1961, 0.9806],
])


input_size, output_size = (2, 3)
input_layer = layers.StepLayer(
    input_size,
    weight=np.array([
        [0.7071, -0.7071],
        [0.7071, 0.7071],
        [-1, 0],
    ]).T,
    function_coef={'lower_value': 0, 'upper_value': 1}
)
output_layer = layers.CompetitiveOutputLayer(output_size)

hamming_network = algorithms.Instar(
    input_layer > output_layer,
    use_bias=False,
    step=0.5,
    n_unconditioned=0
)

with hamming_network as hn:
    hn.train(input_data, epochs=1)
    print(hn.input_layer.weight)



import sys

from neuralpy.helpers.logs import Verbose


__all__ = ('twilio_sms',)


def twilio_sms(account_id, token, to_phone, from_phone, verbose=True):
    """ Send SMS via Twilio service.

    Parameters
    ----------
    account_id : str
        Twilio account ID.
    token : str
        Twilio account token.
    to_phone : str
        SMS receiver phone number.
    from_phone : str
        SMS sender phone number.
    verbose : bool
        Logging verbose. Defaults to ``True``.

    Returns
    -------
    func
        Retunr function which take one text message argument and send it
        via Twilio API.
    """
    logs = Verbose(verbose=verbose).logs

    try:
        import twilio
    except ImportError:
        logs.error("Install twilio module before use this function. Command:")
        logs.simple("pip install twilio")
        sys.exit()

    def send_message(text_message):
        logs.log("Message", "Send SMS with text: '{}'".format(text_message))

        client = twilio.rest.TwilioRestClient(account_id, token)
        message = client.messages.create(body=text_message, to=to_phone,
                                         from_=from_phone)
        return message
    return send_message



import numpy as np


__all__ = ('add_bias_column', 'iter_until_converge', 'shuffle')


def add_bias_column(data):
    """ Transform input data to matrix and add as first column which
    contains all ones.

    Parameters
    ----------
    data : array-like
        Main matrix.

    Returns
    -------
    array-like
        Copied matrix with the new column at first position which contains
        all ones.
    """
    data = np.asmatrix(data)
    bias_vector = np.ones((data.shape[0], 1))
    return np.concatenate((bias_vector, np.asarray(data)), axis=1)


def iter_until_converge(network, epsilon):
    epoch = network.epoch
    error = epsilon + 1

    while np.any(error > epsilon):
        yield epoch
        epoch = epoch + 1
        error = network.last_error()


def shuffle(*arrays):
    if not arrays:
        return tuple()

    first = arrays[0]
    n_samples = first.shape[0]

    for array in arrays:
        if n_samples != array.shape[0]:
            raise ValueError("All matrices muts have the same number of rows")

    indices = np.arange(n_samples)
    np.random.shuffle(indices)

    arrays = list(arrays)
    for i, array in enumerate(arrays):
        arrays[i] = array[indices]

    return arrays



from scipy.optimize import minimize_scalar

from neuralpy.core.properties import NonNegativeNumberProperty, ChoiceProperty
from .base import SingleStep


__all__ = ('LinearSearch',)


class LinearSearch(SingleStep):
    """ Linear search for the step selection. Basicly this algorithms
    try different steps and compute your predicted error, after few
    iteration it will chose one which was better.

    Parameters
    ----------
    tol : float
        Tolerance for termination, default to ``0.3``. Can be any number
        greater that zero.
    search_method : 'gloden', 'brent'
        Linear search method. Can be ``golden`` for golden search or ``brent``
        for Brent's search, default to ``golden``.

    Attributes
    ----------
    {first_step}

    Warns
    -----
    {bp_depending}

    Examples
    --------
    >>> import numpy as np
    >>> np.random.seed(0)
    >>>
    >>> from sklearn import datasets, preprocessing
    >>> from sklearn.cross_validation import train_test_split
    >>> from neuralpy import algorithms, layers
    >>> from neuralpy.functions import rmsle
    >>>
    >>> dataset = datasets.load_boston()
    >>> data, target = dataset.data, dataset.target
    >>>
    >>> data_scaler = preprocessing.MinMaxScaler()
    >>> target_scaler = preprocessing.MinMaxScaler()
    >>>
    >>> x_train, x_test, y_train, y_test = train_test_split(
    ...     data_scaler.fit_transform(data),
    ...     target_scaler.fit_transform(target),
    ...     train_size=0.85
    ... )
    >>>
    >>> cgnet = algorithms.ConjugateGradient(
    ...     connection=[
    ...         layers.SigmoidLayer(13),
    ...         layers.SigmoidLayer(50),
    ...         layers.OutputLayer(1),
    ...     ],
    ...     search_method='golden',
    ...     optimizations=[algorithms.LinearSearch],
    ...     verbose=False
    ... )
    >>>
    >>> cgnet.train(x_train, y_train, epochs=100)
    >>> y_predict = cgnet.predict(x_test)
    >>>
    >>> real = target_scaler.inverse_transform(y_test)
    >>> predicted = target_scaler.inverse_transform(y_predict)
    >>>
    >>> error = rmsle(real, predicted.round(1))
    >>> error
    0.20752676697596578

    See Also
    --------
    :network:`ConjugateGradient`
    """
    tol = NonNegativeNumberProperty(default=0.3)
    search_method = ChoiceProperty(choices=['golden', 'brent'],
                                   default='golden')

    def set_weights(self, new_weights):
        for layer, new_weight in zip(self.train_layers, new_weights):
            layer.weight = new_weight.copy()

    def check_updates(self, new_step, weights, delta):
        self.set_weights(weights)
        self.step = new_step

        super(LinearSearch, self).update_weights(delta)
        predicted_output = self.predict(self.input_train)
        return self.error(predicted_output, self.target_train)

    def update_weights(self, weight_deltas):
        real_weights = [layer.weight for layer in self.train_layers]
        res = minimize_scalar(
            self.check_updates, args=(real_weights, weight_deltas),
            tol=self.tol, method=self.search_method,
            options={'xtol': self.tol}
        )

        self.set_weights(real_weights)
        self.step = res.x

        return super(LinearSearch, self).update_weights(weight_deltas)



from numpy import arccos, dot, reshape
from numpy.linalg import norm

from neuralpy.core.properties import DictProperty
from neuralpy.functions import get_partial_for_func
from neuralpy.functions import (linear, sigmoid, signum, tanh, rectifier,
                                softplus, softmax)
from neuralpy.layers.base import BaseLayer


__all__ = ('Layer', 'LinearLayer', 'SigmoidLayer', 'StepLayer', 'TanhLayer',
           'RectifierLayer', 'SoftplusLayer', 'SoftmaxLayer',
           'EuclideDistanceLayer', 'AngleDistanceLayer')


class Layer(BaseLayer):
    """ Base class for neural network layers.

    Parameters
    ----------
    function_coef : dict
        Default settings for activation function.
    {layer_params}
    """

    function_coef = DictProperty()

    def __init__(self, *args, **kwargs):
        super(Layer, self).__init__(*args, **kwargs)

        if self.function_coef is not None:
            partial_func = get_partial_for_func(self.activation_function)
            self.activation_function = partial_func(
                self.activation_function, **self.function_coef
            )


class LinearLayer(Layer):
    """ Layer with linear activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = linear


class SigmoidLayer(Layer):
    """ Layer with sigmoid activation function.

    Parameters
    ----------
    function_coef : dict
        Default configurations for signoid activation function. There is one
        value name ``alpha`` (default is ``1``). `alpha` control your
        function shape.
    {layer_params}
    """
    function_coef = DictProperty(default={'alpha': 1})
    activation_function = sigmoid


class StepLayer(Layer):
    """ Layer with step activation function.

    Parameters
    ----------
    function_coef : dict
        Default configurations for step activation function. There are two
        values ``upper_value`` and ``lower_value`` which are equal to ``1``
        and ``-1`` respectively.
    {layer_params}
    """
    function_coef = DictProperty(default={'upper_value': 1,
                                          'lower_value': -1})
    activation_function = signum


class TanhLayer(Layer):
    """ Layer with `tanh` activation function.

    Parameters
    ----------
    function_coef : dict
        Default configurations for sigmoid activation function. There is one
        value name ``alpha`` (default is ``1``). `alpha` control your
        function shape.
    {layer_params}
    """
    function_coef = DictProperty(default={'alpha': 1})
    activation_function = tanh


class RectifierLayer(Layer):
    """ Layer with rectifier activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = rectifier


class SoftplusLayer(Layer):
    """ Layer with softplus activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = softplus


class SoftmaxLayer(Layer):
    """ Layer with softmax activation function.

    Parameters
    ----------
    function_coef : dict
        Default configurations for softmax activation function. There is one
        value name ``temp`` (default is ``1``). Smaller ``temp`` value will
        make your winner probability closer to ``1``. To big ``temp`` value
        will make all your probabilities closer to equal values.
    {layer_params}
    """
    function_coef = DictProperty(default={'temp': 1})
    activation_function = softmax


class EuclideDistanceLayer(BaseLayer):
    """ Layer output equal to Euclide distance between input value
    and weights.

    Parameters
    ----------
    {layer_params}
    """
    def output(self, input_value):
        input_data = self.preformat_input(input_value)
        distance = norm(input_data.T - self.weight, axis=0)
        return -reshape(distance, (1, self.weight.shape[1]))


class AngleDistanceLayer(BaseLayer):
    """ Layer which output equal to cosine distance between input value
    and weights.

    Parameters
    ----------
    {layer_params}
    """
    def output(self, input_value):
        input_data = self.preformat_input(input_value)
        norm_prod = norm(input_data) * norm(self.weight, axis=0)
        summated_data = dot(input_data, self.weight)
        return -reshape(arccos(summated_data / norm_prod),
                        (1, self.weight.shape[1]))



from numpy import concatenate, array

from neuralpy.core.properties import NonNegativeIntProperty
from neuralpy.network.learning import SupervisedLearning
from neuralpy.network.connections import FAKE_CONNECTION
from neuralpy.network.base import BaseNetwork


__all__ = ('CMAC',)


class CMAC(SupervisedLearning, BaseNetwork):
    """ CMAC Network based on memory.

    Notes
    -----
    * Network always use Mean Absolute Error (MAE).
    * Works for single and multi output values.

    Parameters
    ----------
    quantization : int
        Network transform every input to discrete values. Quantization
        value contol number of total possible values after
        quantization, defaults to ``10``.
    associative_unit_size : int
        Number of associative blocks in memory, defaults to ``2``.
    {step}
    {show_epoch}
    {shuffle_data}
    {full_signals}

    Methods
    -------
    {fit}
    {supervised_train}
    {predict}
    {last_error}
    {plot_errors}

    Examples
    --------
    >>> import numpy as np
    >>> from neuralpy.algorithms import CMAC
    >>>
    >>> train_space = np.linspace(0, 2 * np.pi, 100)
    >>> test_space = np.linspace(np.pi, 2 * np.pi, 50)
    >>>
    >>> input_train = np.reshape(train_space, (100, 1))
    >>> input_test = np.reshape(test_space, (50, 1))
    >>>
    >>> target_train = np.sin(input_train)
    >>> target_test = np.sin(input_test)
    >>>
    >>> cmac = CMAC(
    ...     quantization=100,
    ...     associative_unit_size=32,
    ...     step=0.2,
    ... )
    ...
    >>> cmac.train(input_train, target_train, epochs=100)
    >>> predicted_test = cmac.predict(input_test)
    >>> cmac.error(target_test, predicted_test)
    0.0023639417543036569
    """
    quantization = NonNegativeIntProperty(default=10)
    associative_unit_size = NonNegativeIntProperty(default=2, min_size=2)

    def __init__(self, **options):
        self.weights = {}
        super(CMAC, self).__init__(FAKE_CONNECTION, **options)

    def setup_defaults(self):
        del self.use_bias
        del self.error
        super(CMAC, self).setup_defaults()

    def predict(self, input_data):
        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        predicted = []

        for input_sample in self.quantize(input_data):
            coords = get_memory_coords(input_sample)
            predicted.append(get_result_by_coords(coords))

        return array(predicted)

    def get_result_by_coords(self, coords):
        return sum(
            self.weights.setdefault(coord, 0) for coord in coords
        ) / self.associative_unit_size

    def get_memory_coords(self, quantized_value):
        assoc_unit_size = self.associative_unit_size

        for i in range(assoc_unit_size):
            point = ((quantized_value + i) / assoc_unit_size).astype(int)
            yield tuple(concatenate([point, [i]]))

    def quantize(self, input_data):
        return (input_data * self.quantization).astype(int)

    def train_epoch(self, input_train, target_train):
        get_memory_coords = self.get_memory_coords
        get_result_by_coords = self.get_result_by_coords
        weights = self.weights

        quantized_input = self.quantize(input_train)
        errors = 0

        for input_sample, target_sample in zip(quantized_input, target_train):
            coords = list(get_memory_coords(input_sample))
            predicted = get_result_by_coords(coords)

            error = target_sample - predicted
            for coord in coords:
                weights[coord] += self.step * error

            errors += abs(error)
        return errors / input_train.shape[0]
